#!/bin/sh
# Copyright (c) 2000-2014 Synology Inc. All rights reserved.

# usage, example

HA_PREFIX="/var/packages/HighAvailability/target"
. $HA_PREFIX/etc/rc.subr

action=""
interface=""
ip=""
netmask=""
port=5000
protocol="http"
C_PING=/bin/ping
C_CAT=/bin/cat
C_RM=/bin/rm
C_WGET=/usr/bin/wget
C_DD=/bin/dd
C_ECHO=/bin/echo
C_MV=/bin/mv
C_IFCON=/sbin/ifconfig
C_SLEEP=/bin/sleep
C_TOUCH=/bin/touch
C_LOGGER=/usr/bin/logger
C_IP=/sbin/ip

ResTimeLimit=2 # ms
speedLimit='50' #50M
speedLimitSecure='25' #25M

TEMPPATH=/tmp
TESTFILE_PATH=/webman/3rdparty/HighAvailability
TESTFILE_PREFIX=/usr/syno/synoman
TESTFILE=10M

TMPFILE=$TEMPPATH/.HA.CheckHeartbeatLink.tmp
CHECKING_FILE=$TEMPPATH/.HA.CHECK_HEARTBEAT_ENV
RESULT_FILE=$TEMPPATH/ha/HA.Heartbeat.check.result
is_active=
auto_reset=0
cmdline=

log_notice()
{
	$C_LOGGER -p user.notice -t "$0[$$]" "[HA-NOTICE] ${1:-}"
}

# $1 number
# $2 operator, such as >, <, ==, ...
# $3 number
compare()
{
	python -c "import sys; sys.exit(not($1 $2 $3));"
}

get_heartbeat_mtu()
{
	heartbeat_mtu=$(/usr/syno/bin/synogetkeyvalue /etc.defaults/synoinfo.conf sha_heartbeat_mtu)
	if [ -n "$heartbeat_mtu" ]; then
		printf "%s" "$heartbeat_mtu"
	else
		printf "2000"
	fi
}

record_header()
{
	{
		$C_ECHO "======================="
		$C_ECHO "==$(date "+%Y/%m/%d %H:%M:%S")=="
		$C_ECHO "======================="
	} >> $RESULT_FILE 2>&1
}

record_content()
{
	content=$1
	$C_CAT "$content" >> $RESULT_FILE 2>&1
	$C_ECHO "" >> $RESULT_FILE 2>&1
}

command_exec_to_tmpfile()
{
	command=$1
	echovalue=${2:-0}

	record_header

	$C_ECHO "$command" > $TMPFILE 2>&1
	if [ 0 -ne "$echovalue" ] ; then
		$command >> $TMPFILE 2>&1 || (log_notice "$command"; $C_ECHO "$echovalue")
	else
		$command >> $TMPFILE 2>&1
	fi

	record_content $TMPFILE
}

# see also HAHeartbeatCheck at synoha/src/lib/ha.cc.
reset_env()
{
	interface=$1

	if [ -z "$interface" ] ; then
		$C_ECHO 9
		exit
	fi
	if [ -e $CHECKING_FILE ] ; then
		$C_RM $CHECKING_FILE 2> /tmp/rmfile
		IFEXIST=$($C_IP -o -f inet addr show | grep "${interface}:DRBD")
		if [ -n "$IFEXIST" ] ; then
			command_exec_to_tmpfile "$C_IFCON ${interface}:DRBD down" 2
		fi
		if [ -z $is_active ] ; then
			command_exec_to_tmpfile "$C_RM $TESTFILE_PREFIX/$TESTFILE_PATH/$TESTFILE" 3
		fi
		if [ $auto_reset -eq 1 ] ; then
			command_exec_to_tmpfile "/var/packages/HighAvailability/target/sbin/synoha --restore-drbd-attr ${interface}"
		fi
		$C_ECHO 1
		/usr/bin/killall "$(basename "$0")"
	fi
}

# see also HAHeartbeatCheck at synoha/src/lib/ha.cc.
prepare_env()
{
	interface=$1
	ip=$2
	netmask=$3

	$C_RM $RESULT_FILE
	if [ -z "$netmask" ] ; then
		echo 9
		exit
	fi
	$C_TOUCH $CHECKING_FILE
	if ! wait_for "is_interface_link_up $interface" 30; then
		log_notice "Error: wait HB interface [$interface] link up timeout"
	fi
	command_exec_to_tmpfile "$C_IFCON $interface:DRBD $ip netmask $netmask" 3
	if [ -z $is_active ] ; then
		command_exec_to_tmpfile "$C_DD if=/dev/zero of=$TEMPPATH/$TESTFILE bs=1M count=10" 4
		command_exec_to_tmpfile "$C_MV $TEMPPATH/$TESTFILE $TESTFILE_PREFIX/$TESTFILE_PATH/$TESTFILE" 5
	fi
	$C_ECHO 1
	auto_reset=1

	# sleep 60 and reset
	{
		$C_SLEEP 60 && reset_env
	}&
}

action=$1; shift
case "$action" in
	prepare_active)
		is_active=1
		prepare_env "$@"
		;;
	reset_active)
		is_active=1
		reset_env "$@"
		;;
	prepare_passive)
		prepare_env "$@"
		exit
		;;
	reset_passive)
		reset_env "$@"
		exit
		;;
	check)
		interface=$1
		ip=$2
		port=$3
		protocol=$4
		if [ -z "$interface" ] ; then
			$C_ECHO 9
			exit
		fi

		# Make sure the network environment is ready for the following tests.
		for i in $(seq 1 1 60)
		do
			command_exec_to_tmpfile "$C_PING $ip -s $(get_heartbeat_mtu) -I $interface -c 1 -W 1"
			lossRate=$(grep packets $TMPFILE | cut -d" " -f 6)
			$C_RM $TMPFILE

			if [ "$lossRate" = "0%" ]; then
				break;
			fi

			if [ 60 -eq "$i" ]; then
				$C_ECHO "2"
				exit;
			fi

			# sleep for 1 sec for each iteration to prevent all
			# iterations get "Destination Host Unreachable" result
			# which will report result immediately rather than
			# wait for the timeout we set (1 sec)
			$C_SLEEP 1
		done

		cmdline="$C_PING $ip -s $(get_heartbeat_mtu) -I $interface -c 5 -W 1"
		# ping more 10 packets for bond which is not stable at up
		if echo "$interface" | grep -q bond; then
			command_exec_to_tmpfile "$cmdline"
			command_exec_to_tmpfile "$cmdline"
		fi

		command_exec_to_tmpfile "$cmdline"
		resTime=$(grep "\/avg\/" $TMPFILE | cut -d"/" -f 5)
		lossRate=$(grep packets $TMPFILE | cut -d" " -f 6)
		$C_RM $TMPFILE

		if [ "$lossRate" = "100%" ]; then
			log_notice "$cmdline"
			$C_ECHO "2"
			exit;
		fi
		if [ "$lossRate" != "0%" ]; then
			log_notice "$cmdline"
			$C_ECHO "3"
			exit;
		fi
		if compare "$resTime" ">" "$ResTimeLimit"; then
			log_notice "$cmdline"
			$C_ECHO "5"
			exit;
		fi

		for times in $(seq 1 10)
		do
			#2014-05-12 18:45:41 (108 MB/s) - '10M' saved [10485760/10485760]
			cmdline="$C_WGET --progress=bar:force --tries=1 --timeout=1 --no-proxy --no-check-certificate --delete-after $protocol://$ip:${port}${TESTFILE_PATH}/$TESTFILE"
			command_exec_to_tmpfile "$cmdline"
			speed=$(grep saved $TMPFILE | cut -d"(" -f 2 | cut -d" " -f 1)
#			echo "Speed: $speed" >> /tmp/.checkResultXHeartBeat
			if [ -z "$speed" ] ; then
				log_notice "$cmdline"
				$C_ECHO "6"
				break;
			fi

			if [ "https" = "$protocol" ]; then
				speedTarget=$speedLimitSecure
			else
				speedTarget=$speedLimit
			fi
			if compare "$speed" ">" "$speedTarget"; then
				break;
			fi

			if [ "$times" = "10" ] ; then
				log_notice "$cmdline"
				$C_ECHO "7"
				break;
			fi

			# CPU may be very busy due to a lot of service will restart many times after
			# net related hooks triggered many times when env checking.
			$C_SLEEP 5
		done

		$C_ECHO "1";

		;;
esac

