#!/bin/sh
#
# Usage:
#        handler_sb.sh safemode [ --boot ]
#        After 5.0 hotfix 1, SHA support safemode to handle split-brain
#
#        handler_sb.sh unbind
#
# return 1: error, unbind local
#        2: standby failed
#
# vim:ft=sh

# see EXAMPLES in `man 1 flock`
# shellcheck disable=SC2015
[ "${FLOCKER}" != "$0" ] && exec env FLOCKER="$0" flock -en "$0" "$0" "$@" || :

HA_PREFIX=/var/packages/HighAvailability/target

. $HA_PREFIX/etc/rc.subr
PATH=/sbin:/bin:/usr/sbin:/usr/bin:/usr/syno/sbin:/usr/syno/bin:/usr/local/sbin:/usr/local/bin:/var/packages/HighAvailability/target/sbin:/var/packages/HighAvailability/target/usr/bin

SYNOVSPACE="/usr/syno/bin/synovspace"

unbind()
{
	synoha_log "SB: Split brain handler failed: $1"
	$SYNOHA_BIN --handle-sb &
	rm -f $HA_DRBD_SPLIT_BRAIN
	exit 1
}

clear_HA_states()
{
	synoha_log notice "SB: Clear HA status"
	crm resource cleanup DRBD > /dev/null 2>&1
	crm -F node clearstate "$LOCAL_HOST" > /dev/null 2>&1
}

send_cib_info()
{
	while :; do
		[ -f "$HA_SAFEMODE_INFO" ] || exit 0

		if ! /bin/ps axw | grep "synohainfo --set-info" | grep -v grep >/dev/null 2>&1; then
			"${HA_SBIN_DIR}/synohainfo" --set-info
		fi

		sleep 30
	done
}

standby_and_set_constraint()
{
	for h in "$NODE_HOST0" "$NODE_HOST1"; do
		# when bootup, only set constraint & standby to LOCAL_HOST
		[ "--boot" = "$hasReboot" ] && [ "$LOCAL_HOST" != "$h" ] && continue

		cons=$(printf "<constraints><rsc_location id='RULE_SB_%s' node='%s' rsc='DUMMY_START' score='-INFINITY'/></constraints>" "$h" "$h")
		cibadmin -Mc --xml-text "$cons"

		stby=$(printf "<instance_attributes id='nodes-%s'><nvpair id='nodes-%s-standby' name='standby' value='on'/></instance_attributes>" "$h" "$h")
		cibadmin -Mc --xml-text "$stby"
	done
}

_check_standby()
{
	_dummy_start_location=$(crm_resource --resource DUMMY_START --locate)
	_infoset_location=$(crm_resource --resource INFO_SET --locate)
	_drbd_location=$(crm_resource --resource DRBD --locate)
	_checkpointlast_location=$(crm_resource --resource CHECKPOINT_LAST --locate)
	_confsync_location=$(crm_resource --resource CONF_SYNC --locate)
	_all_location="$_dummy_start_location $_infoset_location $_drbd_location $_checkpointlast_location $_confsync_location"

	if is_support_xa; then
		_network_location=$(crm_resource --resource NETWORK --locate)
		_all_location="$_all_location $_network_location"
	fi

	_localhost_stby=$(echo "$_all_location" | grep "$LOCAL_HOST")
	_host0_stby=$(echo "$_all_location" | grep "$NODE_HOST0")
	_host1_stby=$(echo "$_all_location" | grep "$NODE_HOST1")

	if [ "--boot" = "$hasReboot" ]; then
		# only wait for local standby after reboot
		[ -z "$_localhost_stby" ]
	else
		[ -z "$_host0_stby" ] && [ -z "$_host1_stby" ]
	fi

	_ret=$?

	if [ 0 -ne $_ret ]; then
		standby_and_set_constraint
	fi

	return $_ret
}

start_handle_sb_main()
{
	hasReboot=$1

	synoha_log warning "SB: Split brain handler"
	touch $HA_DRBD_SPLIT_BRAIN

	[ -n "$NODE_HOST0" ] || unbind "host0 missing"
	[ -n "$NODE_HOST1" ] || unbind "host1 missing"

	mkdir -p $HA_SAFEMODE_DIR
	touch $HA_SAFEMODE_INFO

	synoha --notify safe-mode > /dev/null 2>&1

	[ "--boot" != "$hasReboot" ] && synoha --save-ha-state-time split-brain > /dev/null 2>&1

	rm -f $PREVIOUS_ROLE_ACTIVE
	rm -f $PREVIOUS_ROLE_PASSIVE

	[ "--boot" != "$hasReboot" ] && synoha --remote-run-safemode

	standby_and_set_constraint

	synoha_log warning "SB: Split brain handler done"

	IP_managed=$(crm_resource --resource IP --get-parameter is-managed --meta 2>/dev/null)
	if [ "false" = "$IP_managed" ]; then
		crm_resource --resource IP --set-parameter is-managed --meta --parameter-value true
	fi

	_sleep_sec=5
	if ! HASleepFor _check_standby $MAX_HA_SERV_STOP_SEC $_sleep_sec > /dev/null 2>&1; then
		synoha_log "SB: Standby failed, force reboot"
		crm_mon -n -1 | synoha_log notice
		clear_HA_states
		rm -f $HA_DRBD_SPLIT_BRAIN
		touch $SYNO_HA_FORCE_STOP
		/sbin/reboot -f
		exit 2
	fi

	clear_HA_states

	prepare_spaces

	# Make all shares read-only
	synoha --lock-shares

	# Remove all sessions
	target_sessions="$(jq -rs 'map(select(.duration != "-1") | .id) | join(" ")' /usr/syno/etc/private/session/current.users)"
	for s in $target_sessions; do
		rm -f "/usr/syno/etc/private/session/current.users.access.time/$s"
	done

	ha_safemode_backup_user_settings

	ha_safemode_delete_new_app_notification # After user-settings backed up

	# Start rsync daemon
	/sbin/initctl start pkg-synoha-rsyncd
	/sbin/initctl start pkg-synoha-rsyncd-backup-logs
	/sbin/initctl start pkg-synoha-lsyncd-backup-logs

	# Keep required services
	synoha_log notice "SB: Pause all services by reason ha-split-brain except those in ha-safe-mode"
	$SYNOSERVICECFG_BIN --pause-all ha-split-brain ha-safe-mode
	synoha_log notice "SB: Resume all services paused by reason ha-passive"
	$SYNOSERVICECFG_BIN --resume-all ha-passive
	synoha_log notice "SB: Start ssh-shell"
	$SYNOSERVICECFG_BIN --start ssh-shell
	synoha_log notice "SB: Start pkgctl-FileStation"
	# use synopkg instead of synoservice to handle packages
	${SYNO_PKG_BIN} start FileStation

	/usr/syno/bin/synobootseq --set-boot-done
	send_cib_info &

	# Retreive local shares list
	/usr/syno/bin/synowebapi --exec api=SYNO.Core.Share method=list version=1 2> /dev/null | \
		/bin/jq -c 'if .success == true then .data.shares | map({"name": (.vol_path+"/"+.name)}) else [] end' > "${HA_SAFEMODE_DIR}shares.$(cat $CRM_HOSTNAME_ORIGINAL)"

	ha_safemode_enable_ha_ip_on_left_node

	rm -f $HA_DRBD_SPLIT_BRAIN
	synoha_log warning "SB: enter SHA safemode"
} # end of start_handle_sb_main()

start_handle_sb()
{
	hasReboot=$1

	# prevent multiple handle_sb
	if [ "--boot" != "$hasReboot" ]; then
		[ -f "$HA_SAFEMODE_INFO" ] && exit 0
	fi

	start_handle_sb_main "$@" &
}

[ -e /usr/syno/etc/packages/HighAvailability/ha.conf ] || unbind "ha.conf missing"

action=$1; shift
case "$action" in
	safemode)
		start_handle_sb "$@"
		;;
	unbind)
		unbind
		;;
	*)
		exit $LSB_ERR_ARGS
esac

exit $?

